import datetime
import json
from typing import Tuple, List, Dict, Optional

from bs4 import BeautifulSoup

from .source import Source
from ...exceptions import BadResponseException
from ...models import Novel, Volume, Chapter, Metadata

book_info_url = 'https://www.webnovel.com/book/%s'
chapter_list_url = 'https://www.webnovel.com/go/pcm/chapter/get-chapter-list?_csrfToken={csrf}&bookId={id}&pageIndex=0'


class Webnovel(Source):
    name = 'Webnovel'
    base_urls = ('https://www.webnovel.com',)
    last_updated = datetime.date(2021, 9, 3)

    csrf_token: str = None

    def __init__(self):
        super(Webnovel, self).__init__()

    def get_csrf_token(self) -> str:
        return self.session.cookies.get('_csrfToken', '')

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)
        novel_id = self.parse_novel_url(url)

        toc_response = self.request_get(chapter_list_url.format(csrf=self.get_csrf_token(), id=novel_id))
        data = self.validate(toc_response)['data']

        synopsis = [
            para.strip()
            for para in soup.select_one("div[class*='j_synopsis'] > p").find_all(text=True, recursive=False)
            if para.strip()
        ]

        novel = Novel(
            title=data['bookInfo']['bookName'],
            synopsis=synopsis,
            thumbnail_url=f'https://img.webnovel.com/bookcover/{novel_id}',
            url=f'https://www.webnovel.com/book/{novel_id}',
        )

        writer_elements = soup.select('._mn > address > p > *')
        for i in range(len(writer_elements) // 2):
            label = writer_elements[i * 2].text.strip(': ').lower()
            value = writer_elements[i * 2 + 1].text

            if label == 'author':
                novel.author = value
            else:
                novel.metadata.append(Metadata('author', value, others={'role': label}))

        for a in soup.select('.m-tags a'):
            novel.metadata.append(Metadata('subject', a['title']))

        novel.volumes = self.toc(novel_id, data)

        return novel

    def toc(self, novel_id, data: dict) -> List[Volume]:
        volumes = []
        if 'volumeItems' in data:
            for volume_data in data['volumeItems']:
                volumes.append(self.make_volume(novel_id, volume_data['chapterItems'], volume_data))
        elif 'chapterItems' in data:
            volumes.append(self.make_volume(novel_id, data['chapterItems'], None))

        return volumes

    @staticmethod
    def make_volume(novel_id, chapter_items, volume_data: Optional[dict]) -> Volume:
        volume = Volume(volume_data['volumeId'], volume_data['volumeName']) \
            if volume_data else Volume.default()

        for chapter_data in chapter_items:
            if not chapter_data['isAuth']:
                continue

            chapter = Chapter(
                index=chapter_data['chapterIndex'],
                title=f'{chapter_data["chapterIndex"]} {chapter_data["chapterName"]}',
                url=f'https://www.webnovel.com/book/{novel_id}/{chapter_data["chapterId"]}',
            )

            volume.chapters.append(chapter)

        return volume

    def chapter(self, chapter: Chapter):
        novel_id, chapter_id = self.parse_chapter_url(chapter.url)
        response = self.session.get(
            'https://www.webnovel.com/go/pcm/chapter/getContent',
            params={
                '_csrfToken': self.get_csrf_token(),
                'bookId': novel_id,
                'chapterId': chapter_id
            }
        )

        response = self.validate(response)
        data = response['data']['chapterInfo']

        content = BeautifulSoup(
            '<div><p>' + '</p><p>'.join([para['content'] for para in data['contents']]) + '</p></div>', 'lxml')
        self.clean_contents(content)

        chapter.title = data['chapterName']
        chapter.paragraphs = str(content.select_one('div'))

    @staticmethod
    def validate(response) -> Dict:
        parsed = json.loads(response.content)
        try:
            # code 0 denotes a successful response
            if parsed['code'] != 0:
                raise BadResponseException(parsed['msg'])

        # for instances it doesnt return any data
        except KeyError:
            raise BadResponseException('no data was returned')

        return parsed

    @staticmethod
    def parse_novel_url(novel_url):
        """
        :return: url of novel
        """
        novel_url = novel_url.rstrip('/')
        try:
            return int(novel_url.split('/')[4])
        except ValueError:
            return int(novel_url.split('_')[-1])

    @staticmethod
    def parse_chapter_url(chapter_url) -> Tuple[int, int]:
        """
        :return: novel_id, chapter_id
        """
        pieces = chapter_url.rstrip('/').split('/')
        try:
            return int(pieces[4]), int(pieces[5])
        except ValueError:
            return int(pieces[4].split('_')[-1]), int(pieces[5].split('_')[-1])



from dataclasses import dataclass, field
from typing import List

from .chapter import Chapter


@dataclass
class Volume:
    index: int
    name: str

    chapters: List[Chapter] = field(default_factory=lambda: [])

    @staticmethod
    def default():
        return Volume(-1, '_default')



import datetime
import re
from urllib.parse import urlparse

from .source import Source
from ... import BadResponseException
from ...models import Chapter, Novel


class Spacebattles(Source):
    name = 'SpaceBattles'
    base_urls = ('https://forums.spacebattles.com',)
    last_updated = datetime.date(2021, 9, 9)

    def __init__(self):
        super(Spacebattles, self).__init__()
        self.clear_tags = re.compile(r'(</?div>|<br ?/?>)')

    def novel(self, url: str) -> Novel:
        threadmarks_url = url.rstrip('/') + '/threadmarks/'
        soup = self.get_soup(threadmarks_url)

        author_element = soup.select_one('.username')

        # getting writer profile image
        try:
            author_soup = self.get_soup(self.to_absolute_url(author_element['href']))
            avatar = author_soup.select_one('.avatarWrapper > .avatar')
            thumbnail = self.to_absolute_url(avatar['href'])
        except BadResponseException:
            thumbnail = None

        novel = Novel(
            title=soup.select_one('.p-breadcrumbs > li:last-child').text.strip(),
            author=author_element.text,
            thumbnail_url=thumbnail,
            url=url,
        )

        volume = novel.get_default_volume()
        for i, a in enumerate(soup.select('.block-body--threadmarkBody.is-active > div > div > div:first-child a')):
            chapter = Chapter(
                index=i,
                title=a.text,
                url=self.to_absolute_url(a['href'])
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        parsed_url = urlparse(chapter.url)
        raw_url = f'{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}'

        soup = self.get_soup(raw_url)

        article = soup.select_one(f'.u-anchorTarget#{parsed_url.fragment}').parent
        content = article.select_one('.message-inner .message-userContent .bbWrapper')

        content.smooth()
        content.attrs.clear()

        paragraphs = [t for t in self.clear_tags.sub('', str(content)).splitlines() if t]

        chapter.title = article.select_one('.message-cell--threadmark-header > span').text.strip()
        chapter.paragraphs = '<p>' + '</p><p>'.join(paragraphs) + '</p>'



class SourcesException(Exception):
    """base exception for this package"""


class BadResponseException(SourcesException):
    """thrown when an unexpected response is received"""


class UnknownSourceException(SourcesException):
    """thrown when the url does not correspond to an existing source"""


class UnavailableException(SourcesException):
    """thrown when a function is unavailable"""


class ChapterException(SourcesException):
    """thrown when something unexpected happens during chapter update"""



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class BoxNovel(Source):
    base_urls = ('https://boxnovel.com',)
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        authors = soup.select('.author-content a')
        if len(authors) == 2:
            author = authors[0].text.strip() + ' (' + authors[1].text.strip() + ')'
        else:
            author = authors[0].text.strip()

        novel = Novel(
            title=soup.select_one('.breadcrumb > li:last-child').text.strip(),
            author=author,
            synopsis=[t.strip() for t in soup.select_one('.j_synopsis, #editdescription').text.splitlines() if t.strip()],
            thumbnail_url=soup.select_one('.summary_image img')['src'],
            url=url,
        )

        for a in soup.select('a[href*="genre"][rel="tag"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        soup = self.get_soup(url.rstrip('/') + '/ajax/chapters', 'POST')
        volume = novel.get_default_volume()
        for i, a in enumerate(reversed(soup.select('li.wp-manga-chapter a'))):
            chapter = Chapter(
                index=i,
                title=a.text.strip(),
                url=a['href']
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        contents = soup.select_one('div.text-left')
        for element in contents.select('h1, h2, h3, .code-block, script, .adsbygoogle'):
            element.extract()

        chapter.paragraphs = str(contents)



from typing import List, Tuple
from urllib.parse import urlparse

from .source import Source
from ...models import Chapter, Novel, Metadata


class FanFiction(Source):
    base_urls = ('https://www.fanfiction.net',)
    rejected = 'Has cloudflare bot protection'

    def novel(self, url: str) -> Tuple[Novel, List[Chapter], List[Metadata]]:
        soup = self.get_soup(url)
        metadata = []

        cover = None
        image_element = soup.select_one('#profile_top img.cimage')
        if image_element:
            cover = f'https:{image_element["src"]}'

        novel = Novel(
            title=soup.select_one('#profile_top b.xcontrast_txt, #content b').text.strip(),
            author=soup.select_one('#profile_top, #content').select_one('a[href*="/u/"]').text.strip(),
            synopsis=soup.select_one('#profile_top > div').text.strip(),
            thumbnail_url=cover,
            url=url,
        )

        # metadata
        pre_story_links = soup.select('#pre_story_links a')
        if len(pre_story_links) == 2:
            metadata.append(Metadata(
                namespace='fanfiction',
                name=pre_story_links[0].text.strip(),
                value=pre_story_links[1].text.strip(),
            ))
        else:
            metadata.append(Metadata(
                namespace='fanfiction',
                name='Crossover',
                value=pre_story_links[0].text.rstrip(' Crossover'),
            ))

        id_ = urlparse(url).path.split('/')[2]

        chapters = []
        chapter_select = soup.select_one('#chap_select, select#jump')
        if chapter_select:
            # multi chapter fan-fictions
            for i, option in enumerate(chapter_select.select('option')):
                chapter = Chapter(
                    index=i,
                    title=option.text.strip(),
                    url=f'https://www.fanfiction.net/s/{id_}/{option["value"]}'
                )

                chapters.append(chapter)
        else:
            # single chapter fan-fictions
            chapters.append(
                Chapter(
                    index=0,
                    title=novel.title,
                    url=url,
                )
            )

        return novel, chapters, metadata

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        chapter_select = soup.select_one('#chap_select, select#jump')
        if chapter_select:
            title = chapter_select.find('option', selected=True).text.strip()
        else:
            title = soup.select_one('#profile_top b.xcontrast_txt, #content b').text.strip()

        chapter.title = title
        chapter.paragraphs = str(soup.select_one('#storytext, #storycontent'))



from typing import List

from ..crawler import Crawler
from ...models import Metadata


class MetaSource(Crawler):
    lang = 'en'

    @classmethod
    def of(cls, url: str):
        """
        :return: whether the url is from this source
        """
        return any([url.startswith(base_url) for base_url in cls.base_urls])

    def retrieve(self, url) -> List[Metadata]:
        """
        retrieves metadata from url

        :param url: metadata source
        :return: list of metadata
        """
        raise NotImplementedError



import datetime

from .source import Source
from ...models import Novel, Chapter, Metadata


class RoyalRoad(Source):
    name = 'Royal Road'
    base_urls = ('https://www.royalroad.com',)
    last_updated = datetime.date(2021, 9, 6)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('h1[property="name"]').text.strip(),
            author=soup.select_one('span[property="name"]').text.strip(),
            thumbnail_url=soup.select_one('.page-content-inner .thumbnail')['src'],
            synopsis=[p.text.strip() for p in soup.select('.description > [property="description"] > p')],
            url=url,
        )

        for a in soup.select('a.label[href*="tag"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for a in soup.find('tbody').findAll('a', href=True):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=self.base_urls[0] + a["href"],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        contents = soup.select_one('.chapter-content')
        self.clean_contents(contents)

        chapter.title = soup.find('h1').text.strip()
        chapter.paragraphs = str(contents)



import datetime
import re
from typing import List

from .source import Source
from ...models import Novel, Volume, Chapter, Metadata


class WuxiaCo(Source):
    name = 'WuxiaWorld.co'
    base_urls = ('https://www.wuxiaworld.co',)
    last_updated = datetime.date(2021, 9, 4)

    @classmethod
    def of(cls, url: str) -> bool:
        return url.startswith(cls.base_urls[0] + '/')

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.book-name').text.strip(),
            author=soup.select_one('.author span.name').text.strip(),
            synopsis=[p.text.strip() for p in soup.select('div.synopsis p')],
            thumbnail_url=soup.select_one('.book-img img')['src'],
            url=url
        )

        catalog = soup.select_one('.book-catalog .txt')
        if catalog:
            novel.metadata.append(Metadata('subject', catalog.text.strip()))

        volume = Volume.default()
        novel.volumes.append(volume)
        for i, item in enumerate(soup.select('a.chapter-item')):
            # wuxiaco uses inline styling(color) to show that chapter isn't ready yet
            # no need to download chapters without actual content
            if 'style' in item.attrs.keys():
                continue

            title = item.find('p').text

            chapter = Chapter(
                index=i,
                title=title,
                url=self.base_urls[0] + item['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        title = soup.find('h1', {'class': 'chapter-title'}).text

        # remove google ads
        for element in soup.find_all('ins', {'class': 'adsbygoogle'}):
            element.decompose()

        content = [
            text.strip()
            for text in soup.find('div', {'class': 'chapter-entity'}).find_all(text=True, recursive=False)
        ]

        # removes random junk inside paragraphs
        # including new-line, tab
        #
        content = self._clean_content(content)

        chapter.title = title
        chapter.paragraphs = '<p>' + '</p><p>'.join(content) + '</p>'

    def _clean_content(self, content) -> List[str]:
        paragraphs = []

        # check is paragraph has line breaks or tabs inside
        # if split to separate paragraphs
        for i, para in enumerate(content):
            para = re.sub(r'[\n\t\r]', ' ', para)

            parts = [part.strip() for part in para.split('  ') if part]

            # filter out junk lines
            p = ' '.join((
                part
                for part in parts
                if part not in ['Please go to', 'to read the latest chapters for free']
            ))

            paragraphs.append(p)

        return paragraphs



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class NovelsRock(Source):
    name = 'Novels Rock'
    base_urls = ('https://novelsrock.com',)
    last_updated = datetime.date(2021, 9, 4)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.breadcrumb > li:last-child').text.strip(),
            author=', '.join([a.text.strip() for a in soup.select('.author-content > a')]),
            thumbnail_url=soup.select_one('.summary_image img')['src'],
            synopsis=[str(p).strip() for p in soup.select_one('.summary__content').find_all(text=True, recursive=False)],
            url=url,
        )

        for a in soup.select('.summary_content_wrap a[href*="genre"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for a in reversed(soup.select('.wp-manga-chapter > a')):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        contents = soup.select('div.reading-content p')

        body = []
        for p in contents:
            for ad in p.select('h3, .code-block, .adsense-code'):
                ad.decompose()
            body.append(str(p))

            chapter.paragraphs = ''.join(body)



import datetime
import re

from .source import Source
from ...models import Novel, Chapter, Metadata

query = """query chapters_NovelRefetchQuery(
    $slug: String!
    $startChapNum: Int
) {...chapters_list_items} fragment chapters_list_items on Query {
    chapterListChunks(
        bookSlug: $slug,
        chunkSize: %d,
        startChapNum: $startChapNum
    ) {
        items {
            title
            chapNum
            url
        }
    }
}
"""


class NovelFun(Source):
    base_urls = ('https://novelfun.net',)
    last_updated = datetime.date(2021, 9, 7)

    graphql = f'{base_urls[0]}/graphql'

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('h1').text.strip(),
            author=soup.select_one('a[href*="author"]').text.strip(),
            synopsis=[p.text.strip() for p in soup.select('article > div > p')],
            url=url,
        )

        novel.thumbnail_url = soup.select_one(f'img[title="{novel.title}"]')['src']

        for a in soup.select('td a[href*="genre"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        title_slug = url.rstrip('/').split('/')[-1]
        total = int(re.match(r'(\d+)', soup.select_one('h2 ~ div').text).group(1))
        data = {
            'id_': 'chapters_NovelRefetchQuery',
            'query': query % total,
            'variables': {
                'slug': title_slug,
                'startChapNum': 1,
            },
        }
        data = self.session.get(self.graphql, json=data).json()

        novel.get_default_volume().chapters = [
            Chapter(
                index=c['chapNum'],
                title=f'Chapter {c["chapNum"]}: {c["title"]}',
                url=self.base_urls[0]+c['url'],
            )
            for c in data['data']['chapterListChunks'][0]['items']
        ]

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('h1 ~ div')
        self.clean_contents(content)

        chapter.paragraphs = str(content)



import datetime

from .source import Source
from ...models import Novel, Chapter


class WuxiaOnline(Source):
    name = 'WuxiaWorld.online'
    base_urls = ('https://wuxiaworld.online',)
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        synopsis = []
        for element in soup.select('#noidungm > *'):
            if element.name == 'hr':
                break
            elif element.name == 'p':
                if 'style' in element.attrs:
                    continue

                synopsis.append(element.text.strip())

        novel = Novel(
            title=soup.select_one('h1.entry-title').text.strip(),
            thumbnail_url=self.base_urls[0] + soup.select_one('.info_image img')['src'],
            synopsis=synopsis,
            url=url,
        )

        author_element = soup.select_one('a[href*="author"]')
        if author_element:
            novel.author = author_element.text.strip()

        volume = novel.get_default_volume()
        for a in reversed(soup.select('.chapter-list > .row a')):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href']
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('#list_chapter .content-area')

        for h2 in content.select('h2[style="font-weight:bold"]'):
            h2.extract()

        for hidden in content.select('[style="display:none"]'):
            hidden.extract()

        first_element = next(content.children)
        if first_element.name == 'br':
            first_element.extract()

        self.clean_contents(content)

        chapter.paragraphs = str(content)

    def clean_element(self, element):

        try:
            # hidden anti-scraping content
            if 'display:none' in element['style']:
                element.extract()
                return
        except KeyError:
            pass

        try:
            if element.name == 'h2' and 'font-weight:bold' in element['style']:
                next_sibling = element.nextSibling()
                if next_sibling.name == 'br':
                    next_sibling.extract()

                element.extract()
                return
        except KeyError:
            pass

        super(WuxiaOnline, self).clean_element(element)



__version__ = '0.2.0'

from .exceptions import (
    SourcesException, BadResponseException, UnknownSourceException,
    UnavailableException, ChapterException,
)
from .models import Novel, Volume, Chapter, Metadata
from .sources import sources, meta_sources
from .utils.parsers import parse_source, parse_metasource



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class PeachPitting(Source):
    name = 'Peach Pits'
    base_urls = ('https://peachpitting.com',)
    last_updated = datetime.date(2021, 9, 4)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        details_parent, synopsis_parent, *_ = soup.select('.elementor-text-editor')

        novel = Novel(
            title=soup.select_one('h3.elementor-heading-title').text.strip(),
            thumbnail_url=soup.select_one('.elementor-image img')['src'],
            synopsis=[p.text.strip() for p in synopsis_parent.select('p')],
            url=url,
        )

        for p in details_parent.select('p'):
            text = p.text.strip()
            if text.startswith('Author: '):
                novel.author = text.strip('Author: ')
            elif text.startswith('Tags: '):
                for text in text.strip('Tags: ').split(','):
                    novel.metadata.append(Metadata('subject', text.strip()))

        # other contributors
        for a in soup.select('.pp-multiple-authors-wrapper .author'):
            novel.metadata.append(Metadata('contributor', a.text.strip()))

        volume = novel.get_default_volume()
        for a in soup.select('.pt-cv-page > div a'):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('#wtr-content')

        # remove ads
        for div in content.select('.ad-section'):
            div.extract()

        self.clean_contents(content)

        chapter.title = soup.select_one('.post-title').text.strip()
        chapter.paragraphs = str(content)



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class DragonTea(Source):
    name = 'Dragon Tea'
    base_urls = ('https://dragontea.ink/',)
    last_updated = datetime.date(2021, 9, 7)

    bad_tags = [
        'noscript', 'script', 'iframe', 'form', 'hr', 'img', 'ins',
        'button', 'input', 'amp-auto-ads', 'pirate',
        'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
    ]

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.post-title').text.strip(),
            author=soup.select_one('.author-content').text.strip(),
            thumbnail_url=soup.select_one('.summary_image img')['src'],
            synopsis=[p.text.strip() for p in soup.select('.summary__content > p')],
            url=url,
        )

        # other metadata
        for item in soup.select('.post-content_item'):
            key = item.select_one('.summary-heading').text.strip()
            value = item.select_one('.summary-content').text.strip()
            if key == 'Alternative':
                novel.metadata.append(Metadata('title', value, others={'role': 'alt'}))
            elif key == 'Type':
                novel.metadata.append(Metadata('type', value))

        for a in soup.select('.genres-content > a'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        for a in soup.select('.tags-content > a'):
            novel.metadata.append(Metadata('tag', a.text.strip()))

        artist_content = soup.select_one('.artist-content > a')
        if artist_content:
            novel.metadata.append(Metadata('contributor', artist_content.text.strip(), others={'role': 'ill'}))

        soup = self.get_soup(url.rstrip('/') + '/ajax/chapters/', method='POST')
        volume = novel.get_default_volume()
        for a in reversed(soup.select('.wp-manga-chapter > a')):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.reading-content')

        # text-left has a better collection of paragraphs...
        # however we are not taking any chances assuming its always there
        text_left = content.select_one('.text-left')
        if text_left:
            content = text_left

        self.clean_contents(content)

        chapter.title = soup.select_one('.breadcrumb > li.active').text.strip()
        chapter.paragraphs = str(content)



import datetime
from typing import List

import requests

from .metasource import MetaSource
from ...models import Metadata


class WlnUpdates(MetaSource):
    base_urls = ('https://www.wlnupdates.com/',)
    last_updated = datetime.date(2021, 8, 25)

    api_endpoint = 'https://www.wlnupdates.com/api'

    def retrieve(self, url) -> List[Metadata]:
        metadata = []
        
        # get json data
        id = int(url.split('/')[4])
        data = self.api_request(id, 'get-series-id')['data']

        # alternate names
        for name in data['alternatenames']:
            if name == data['title']:
                continue

            metadata.append(Metadata('title', name, others={'type': 'alternate'}))

        # illustrators
        for obj in data['illustrators']:
            metadata.append(Metadata('contributor', obj['illustrator'], others={'role': 'ill'}))

        # publishers
        for obj in data['publishers']:
            metadata.append(Metadata('publisher', obj['publisher']))

        # genre
        for obj in data['genres']:
            metadata.append(Metadata('subject', obj['genre']))

        # tags
        for obj in data['tags']:
            metadata.append(Metadata('tag', obj['tag']))

        # original language
        if data['orig_lang']:
            metadata.append(Metadata('lang', data['orig_lang'], others={'id_': 'original language'}))

        # publication
        if data['pub_date']:
            metadata.append(Metadata('date', data['pub_date'], others={'role': 'publication'}))

        return metadata

    def api_request(self, id: int, mode: str):
        return self.session.post(self.api_endpoint, json={'id': id, 'mode': mode}).json()



import datetime
import re
from time import time
from urllib.parse import urlparse

from .source import Source
from ...models import Chapter, Novel, Metadata

chapter_info_url = 'https://www.wattpad.com/v4/parts/{id}?fields=id,title,pages,text_url&_={key}'
story_info_url = 'https://www.wattpad.com/api/v3/stories/{id}'


class WattPad(Source):
    name = 'Wattpad'
    base_urls = ('https://www.wattpad.com', 'https://my.w.tt',)
    last_updated = datetime.date(2021, 9, 6)

    def __init__(self):
        super(WattPad, self).__init__()
        self.decimal = re.compile(r'\d+')

    def novel(self, url: str) -> Novel:
        info_url = story_info_url.format(id=self.decimal.search(url).group())
        data = self.request_get(info_url).json()

        novel = Novel(
            title=data['title'],
            thumbnail_url=data['cover'],
            author=data['user']['name'],
            synopsis=[t.strip() for t in data['description'].split('\n') if t.strip()],
            url=url,
        )

        for tag in data['tags']:
            novel.metadata.append(Metadata('tag', tag))

        novel.metadata.append(Metadata('status', 'Completed' if data['completed'] else 'Ongoing'))
        novel.metadata.append(Metadata('date', data['createDate']))

        volume = novel.get_default_volume()
        for part in data['parts']:
            chapter = Chapter(
                index=len(volume.chapters),
                title=part['title'],
                url=part['url'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        chapter_id = urlparse(chapter.url).path.split('-', maxsplit=1)[0].strip('/')
        info_url = chapter_info_url.format(id=chapter_id, key=int(time() * 1000))
        data = self.request_get(info_url).json()

        text = self.request_get(data['text_url']['text']).content.decode('utf-8')
        text = re.sub(r'<p data-p-id="[a-f0-9]+">', '<p>', text)

        chapter.title = data['title']
        chapter.paragraphs = text



import datetime

from .source import Source
from ...models import Novel, Chapter, Metadata


class ChickEngege(Source):
    base_urls = ('https://www.chickengege.org',)
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        thumbnail_element = soup.select_one('.novelist-cover-image')

        # remove 'synopsis:' from synopsis paragraphs
        soup.select_one('.novelist-align ~ p strong').extract()

        novel = Novel(
            title=soup.select_one('.entry-title').text.strip(),
            thumbnail_url=thumbnail_element['src'] if thumbnail_element else None,
            synopsis=[p.text.strip() for p in soup.select('.novelist-align ~ p')],
            url=url,
        )

        for element in soup.select('.entry-content > div strong'):
            text = element.text.strip()
            if text == 'Original Title:':
                novel.metadata.append(
                    Metadata('title', element.findNext('span').text.strip(), others={'type': 'original'}))
            elif text == 'Translator(s):':
                ul = element.findNext('ul')
                for a in ul.select('li > a'):
                    novel.metadata.append(Metadata('contributor', a.text.strip(), others={'role': 'translator'}))
            elif text == 'Editor(s):':
                ul = element.findNext('ul')
                for a in ul.select('li > a'):
                    novel.metadata.append(Metadata('contributor', a.text.strip(), others={'role': 'editor'}))

        for a in soup.select('a[rel="tag"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for a in soup.select('#novelList > li > a'):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.entry-wrap .entry-content')
        self.clean_contents(content)

        chapter.title = soup.select_one('.entry-title').text.strip()
        chapter.paragraphs = str(content)



from .betwixtedbutterfly import BetwixtedButterfly
from .boxnovel_com import BoxNovel
from .chickengege import ChickEngege
from .chrysagarden import Chrysanthemumgarden
from .creativenovels import CreativeNovels
from .dragontea import DragonTea
from .dummynovels import DummyNovels
from .fanfiction import FanFiction
from .foxaholic import Foxaholic
from .lightnovelworld import LightNovelWorld
from .mtlnovel import MtlNovel
from .novelfull import NovelFull
from .novelfun import NovelFun
from .novelgate import NovelGate
from .novelhall import NovelHall
from .novelonlinefull import NovelOnlineFull
from .novelpassion import NovelPassion
from .novelsite import NovelSite
from .novelsrock import NovelsRock
from .peachpitting import PeachPitting
from .rainofsnow import RainOfSnow
from .readlightnovel_me import ReadLightNovel
from .readlightnovels_net import ReadLightNovelsNet
from .royalroad import RoyalRoad
from .scribblehub import ScribbleHub
from .spacebattles import Spacebattles
from .sufficientvelocity import SufficientVelocity
from .wattpad import WattPad
from .webnovel import Webnovel
from .wuxiaco import WuxiaCo
from .wuxiacom import WuxiaCom
from .wuxiaonline import WuxiaOnline
from .wuxiasite import WuxiaSite

sources = [
    BetwixtedButterfly,
    BoxNovel,
    ChickEngege,
    Chrysanthemumgarden,
    CreativeNovels,
    DragonTea,
    DummyNovels,
    FanFiction,
    Foxaholic,
    LightNovelWorld,
    MtlNovel,
    NovelFull,
    NovelFun,
    NovelGate,
    NovelHall,
    NovelOnlineFull,
    NovelPassion,
    NovelSite,
    NovelsRock,
    PeachPitting,
    RainOfSnow,
    ReadLightNovel,
    ReadLightNovelsNet,
    RoyalRoad,
    ScribbleHub,
    Spacebattles,
    SufficientVelocity,
    WattPad,
    Webnovel,
    WuxiaCo,
    WuxiaCom,
    WuxiaOnline,
    WuxiaSite,
]



import datetime
import re
from abc import ABC
from typing import List
from urllib.parse import urlparse

import cloudscraper
from bs4 import BeautifulSoup, Comment
from requests.cookies import RequestsCookieJar

from ..exceptions import BadResponseException


class Crawler(ABC):
    lang: str
    base_urls: List[str]
    last_updated: datetime.date
    rejected: str

    def __init__(self):
        self.session = cloudscraper.create_scraper()

    def set_cookies(self, cookies: RequestsCookieJar):
        self.session.cookies = cookies

    def get_soup(self, url: str, method: str = 'GET', **kwargs) -> BeautifulSoup:
        """Download website html and create a bs4 object"""
        soup = BeautifulSoup(self.request(method, url, **kwargs).content, 'lxml')
        if not soup.find('body'):
            raise ConnectionError('HTML document was not loaded correctly.')

        return soup

    def request(self, method: str, url: str, **kwargs):
        """Create a request to the url and return the response if ok

        :raises BadResponseException: if the response is not valid (status_code==200)
        """
        response = self.session.request(method, url, **kwargs)
        if not response.ok:
            raise BadResponseException(response)

        return response

    def request_get(self, url, **kwargs):
        """Creates a get request to the specified url"""
        return self.request('GET', url, **kwargs)

    # ---- Inspired from https://github.com/dipu-bd/lightnovel-crawler ----
    # ----      And almost a perfect copy of the functions below       ----

    bad_tags = [
        'noscript', 'script', 'style', 'iframe', 'ins', 'header', 'footer',
        'button', 'input', 'amp-auto-ads', 'pirate', 'figcaption', 'address',
        'tfoot', 'object', 'video', 'audio', 'source', 'nav', 'output', 'select',
        'textarea', 'form', 'map',
    ]

    blacklist_patterns = []

    notext_tags = [
        'img',
    ]

    preserve_attrs = [
        'href', 'src', 'alt',
    ]

    def is_blacklisted(self, text):
        """
        :return: whether the text is black listed
        """
        for pattern in self.blacklist_patterns:
            if re.search(pattern, text, re.IGNORECASE):
                return True

        return False

    def clean_contents(self, contents):
        if not contents:
            return contents

        contents.attrs = {}
        for element in contents.find_all(True):
            self.clean_element(element)

        return contents

    def clean_element(self, element):
        # remove comments
        if isinstance(element, Comment):
            element.extract()

        elif element.name == 'br':
            next_element = getattr(element, 'next_sibling')
            if next_element and next_element.name == 'br':
                element.extract()

        # Remove bad tags
        elif element.name in self.bad_tags:
            element.extract()

        # Remove empty elements
        elif not element.text.strip():
            if element.name not in self.notext_tags and not element.find_all(recursive=False):
                element.extract()

        # Remove blacklisted elements
        elif self.is_blacklisted(element.text):
            element.extract()

        # Remove attributes
        elif hasattr(element, 'attrs'):
            element.attrs = {key: element.get(key) for key in self.preserve_attrs if key in element.attrs}

    @staticmethod
    def find_paragraphs(element, **kwargs) -> List[str]:
        paragraphs = []
        for t in element.find_all(text=True, **kwargs):
            text = str(t).strip()
            if not text:
                continue

            paragraphs.append(text)

        return paragraphs

    def to_absolute_url(self, url: str, current_url: str = None) -> str:
        if url.startswith('http://') or url.startswith('https:'):
            return url
        if url.startswith('//'):
            return f'{urlparse(current_url or self.base_urls[0]).scheme}:{url}'
        elif url.startswith('/'):
            return self.base_urls[0].lstrip('/') + url

        return current_url.rstrip('/') + url



import datetime
import re
from typing import List, Tuple

from .source import Source
from ...models import Chapter, Novel, Metadata


class Chrysanthemumgarden(Source):
    name = 'Chrysanthemum Garden'
    base_urls = ('https://chrysanthemumgarden.com/',)
    last_updated = datetime.date(2021, 9, 7)

    bad_tags = [
        'div', 'pirate', 'script',
        'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
    ]

    blacklist_patterns = [
        r'^(\s| )+$',  # non-breaking whitespace
    ]

    def novel(self, url: str) -> Tuple[Novel, List[Chapter], List[Metadata]]:
        soup = self.get_soup(url)

        synopsis = []
        synopsis_elements = soup.select('.entry-content > p, hr')
        for element in synopsis_elements:
            if element.tag == 'hr':
                break

            synopsis.append(element.text.strip())

        novel = Novel(
            title=soup.select_one('.novel-title').find(text=True).strip(),
            author=soup.select_one('.author-name > a').text.strip(),
            synopsis=synopsis,
            thumbnail_url=soup.select_one('.novel-cover img')['src'],
            url=url,
        )

        raw_title = soup.select_one('.novel-raw-title')
        if raw_title:
            novel.metadata.append(Metadata('title', raw_title.text.strip(), others={'role': 'alt'}))

        for a in soup.select('.series-tag'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for i, a in enumerate(soup.select('.chapter-item > a')):
            chapter = Chapter(
                index=i,
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('#novel-content')
        self.clean_contents(content)

        chapter.title = soup.select_one('.chapter-title').text.strip()
        chapter.paragraphs = str(content)

    def clean_element(self, element):

        # this removes hidden elements
        try:
            if re.match(r'(height:\s*1px|width:\s*0)', element['style'], flags=re.IGNORECASE):
                element.extract()
                return
        except KeyError:
            pass

        if 'class' in element.attrs and 'jum' in element['class']:
            self.reorder_text(element)

        super(Chrysanthemumgarden, self).clean_element(element)

    def reorder_text(self, element):
        text = ''
        for char in element.text:
            try:
                text += self.jumble_map[char]
            except KeyError:
                text += char

        element.string = text

    jumble_map = {'j': 'a', 'y': 'b', 'm': 'c', 'v': 'd', 'f': 'e', 'o': 'f', 'u': 'g', 't': 'h', 'l': 'i', 'p': 'j',
                  'x': 'k', 'i': 'l', 'w': 'm', 'c': 'n', 'b': 'o', 'q': 'p', 'd': 'q', 'g': 'r', 'r': 's', 'a': 't',
                  'e': 'u', 'n': 'v', 'k': 'w', 'z': 'x', 's': 'y', 'h': 'z', 'C': 'A', 'D': 'B', 'J': 'C', 'G': 'D',
                  'S': 'E', 'M': 'F', 'X': 'G', 'L': 'H', 'P': 'I', 'A': 'J', 'B': 'K', 'O': 'L', 'Z': 'M', 'R': 'N',
                  'Y': 'O', 'U': 'P', 'H': 'Q', 'E': 'R', 'V': 'S', 'K': 'T', 'F': 'U', 'N': 'V', 'Q': 'W', 'W': 'X',
                  'T': 'Y', 'I': 'Z'}



import datetime
from typing import List

from .metasource import MetaSource
from ...models import Metadata


class NovelUpdates(MetaSource):
    base_urls = ('https://www.novelupdates.com',)
    last_updated = datetime.date(2021, 9, 3)

    def retrieve(self, url) -> List[Metadata]:
        metadata = []
        soup = self.get_soup(url)

        # alternate titles
        for text in soup.select_one('#editassociated').find_all(text=True, recursive=False):
            metadata.append(Metadata('title', text.strip(), {'type': 'alternate'}))

        # novel type
        metadata.append(Metadata('type', soup.select_one('.genre.type').text))

        # genre
        for a in soup.select('#seriesgenre > a'):
            metadata.append(Metadata('subject', a.text.strip()))

        # tag
        for a in soup.select('#showtags > a'):
            metadata.append(Metadata('tag', a.text.strip()))

        # og language
        metadata.append(Metadata('lang', soup.select_one('#showlang > a').text, {'id': 'original'}))

        # illustrators
        for a in soup.select('#showartists > a'):
            metadata.append(Metadata('contributor', a.text, {'role': 'ill'}))

        # publishers
        if soup.select_one('#showopublisher > a'):
            metadata.append(Metadata('publisher', soup.select_one('#showopublisher > a').text, {'role': 'original'}))
        if soup.select_one('#showepublisher > a'):
            metadata.append(Metadata('publisher', soup.select_one('#showepublisher > a').text, {'role': 'english'}))

        # publication
        if not soup.select_one('#edityear > .seriesna'):
            metadata.append(Metadata('date', soup.select_one('#edityear').text.strip(), {'role': 'publication'}))

        return metadata



import datetime
from typing import List, Tuple

from bs4 import BeautifulSoup

from .source import Source
from ...models import Chapter, Volume, Novel, Metadata
from ...exceptions import ChapterException


class BetwixtedButterfly(Source):
    name = 'Betwixted Butterfly'
    base_urls = ('https://betwixtedbutterfly.com',)
    last_updated = datetime.date(2021, 9, 3)

    bad_tags = [
        'noscript', 'script', 'iframe', 'form', 'img', 'ins',
        'button', 'input', 'amp-auto-ads', 'pirate',
        'nav', 'h1', 'h2',
    ]

    def novel(self, url: str) -> Tuple[Novel, List[Chapter], List[Metadata]]:
        soup = self.get_soup(url)

        details_parent, synopsis_parent, *_ = soup.select('.elementor-text-editor')

        author = None
        for p in details_parent.select('p'):
            if p.text.startswith('Author: '):
                author = p.text.strip('Author: ')
                break

        novel = Novel(
            title=soup.select_one('h2.elementor-heading-title').text.strip(),
            author=author,
            synopsis=[p.text.strip() for p in synopsis_parent.select('p')],
            thumbnail_url=soup.select_one('.elementor-image img')['src'],
            url=url,
        )

        # tags
        for a in soup.select('a.elementor-button'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = Volume.default()
        novel.volumes.append(volume)
        for a in soup.select('.elementor-tab-content > p a'):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href']
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        elements = soup.select(".entry-inner section .elementor-element:not(.elementor-widget-button, "
                               ".elementor-column)")
        if elements:
            content = BeautifulSoup(f'<div>{"".join([str(element) for element in elements])}</div>', 'lxml').select_one('div')
        else:
            content = soup.select_one('.entry-inner')

        # this source lists chapters that aren't completed/translated in table of contents
        # links of said chapter are redirected to another url that indicates the lack of the chapter
        if content is None:
            raise ChapterException('Absent', f'"{chapter}" redirected to placeholder')

        # remove navigation buttons
        for button in content.select('.elementor-widget-button'):
            button.extract()

        # wp-block-columns -> navigation
        # code-block -> ads
        for div in content.select('.wp-block-columns, .code-block'):
            div.extract()

        self.clean_contents(content)

        chapter.paragraphs = str(content)



from dataclasses import dataclass


@dataclass
class Chapter:
    index: int = -1
    no: float = None
    title: str = None
    paragraphs: str = None
    url: str = None

    def __hash__(self):
        return hash(self.url)

    def __eq__(self, other):
        return self.url == other.url



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class MtlNovel(Source):
    name = 'MTL Novel'
    base_urls = (
        'https://www.mtlnovel.com',
        'http://www.mtlnovel.com',
    )
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Novel:
        http_url = url.replace('https://', 'http://')
        soup = self.get_soup(http_url)

        author = ''
        info_element = soup.select_one('.info > tbody')
        for element in info_element.select('tr'):
            text = element.text
            if text.startswith('Author:'):
                author = text[7:]
                break
        
        novel = Novel(
            title=soup.select_one('.entry-title').text.strip(),
            author=author,
            synopsis=[p.text.strip() for p in soup.select('.desc > p:not(.descr)')],
            thumbnail_url=soup.select_one('.main-tmb')['src'],
            url=url,
        )

        # metadata
        alt_title = soup.select_one('#alt')
        if alt_title:
            novel.metadata.append(Metadata('title', alt_title.text.strip(), others={'role': 'alt'}))

        for genre in soup.select('#genre a'):
            novel.metadata.append(Metadata('subject', genre.text.strip()))

        for tag in soup.select('#tags a:not(.edit-history-button)'):
            novel.metadata.append(Metadata('tag', tag.text.strip()))

        novel.metadata.append(Metadata('status', soup.select_one('#status').text.strip()))

        # chapter list
        chapter_list_url = http_url.rstrip('/') + '/chapter-list/'
        soup = self.get_soup(chapter_list_url)

        volume = novel.get_default_volume()
        for i, a in enumerate(reversed(soup.select('.ch-list .ch-link'))):

            chapter = Chapter(
                index=i,
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        paragraphs = soup.select_one('.par')
        self.clean_contents(paragraphs)

        # remove ad frames
        for frame in paragraphs.select('amp-iframe, .ads'):
            frame.decompose()

        chapter.title = soup.select_one('.current-crumb').text.strip()
        chapter.paragraphs = str(paragraphs)



import datetime
from typing import List, Tuple

from bs4 import BeautifulSoup

from .source import Source
from ...models import Chapter, Volume, Novel, Metadata


class Foxaholic(Source):
    base_urls = ('https://www.foxaholic.com',)
    last_updated = datetime.date(2021, 9, 3)

    bad_tags = [
        'noscript', 'script', 'iframe', 'form', 'hr', 'img', 'ins',
        'button', 'input', 'amp-auto-ads', 'pirate',
        'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
    ]

    blacklist_patterns = [
        r'^[\W\D]*(volume|chapter)[\W\D]+\d+[\W\D]*$',
        r'^ $',  # non-breaking whitespace
    ]

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        hot = soup.select_one('.post-title .hot')
        if hot:
            hot.extract()

        authors = soup.select('.author-content > a')
        synopsis_paras = soup.select('.summary__content > p')

        novel = Novel(
            title=soup.select_one('.post-title').text.strip(),
            author=authors[0].text.strip(),
            thumbnail_url=soup.select_one('.summary_image img')['src'],
            synopsis=[p.text.strip() for p in synopsis_paras],
            url=url,
        )

        # other metadata
        for item in soup.select('.post-content_item'):
            if item.select_one('.summary-heading').text.strip() == 'Alternative':
                for alt in item.select_one('.summary-content').text.strip().split(', '):
                    if alt != novel.title:
                        novel.metadata.append(Metadata('title', alt, others={'role': 'alt'}))
                break

        for author in authors[1:]:
            novel.metadata.append(Metadata('contributor', author.text.strip(), others={'role': 'aut'}))

        for artist in soup.select('.artist-content a'):
            novel.metadata.append(Metadata('contributor', artist.text.strip()))

        for genre in soup.select('.genres-content > a'):
            novel.metadata.append(Metadata('subject', genre.text.strip()))

        for tag in soup.select('.tags-content > a'):
            novel.metadata.append(Metadata('tag', tag.text.strip()))

        novel_id = soup.select_one('.wp-manga-action-button')['data-post']
        response = self.session.post(
            'https://www.foxaholic.com/wp-admin/admin-ajax.php',
            data={
                'action': 'manga_get_chapters',
                'manga': novel_id,
            }
        )

        soup = BeautifulSoup(response.content, 'lxml')

        volume = Volume.default()
        novel.volumes.append(volume)
        for a in reversed(soup.select('.wp-manga-chapter > a')):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.reading-content')

        # text-left has a better collection of paragraphs...
        # however we are not taking any chances assuming its always there
        text_left = content.select_one('.text-left')
        if text_left:
            content = text_left

        for hidden in soup.select('[class*="foxaholic-publift"]'):
            hidden.extract()

        self.clean_contents(content)

        chapter.title = soup.select_one('.breadcrumb > li.active').text.strip()
        chapter.paragraphs = str(content)



import datetime
import re

from .source import Source
from ...models import Chapter, Novel, Metadata


class NovelHall(Source):
    base_urls = ('https://www.novelhall.com',)
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        synopsis_element = soup.select_one('.js-close-wrap')
        synopsis_element.select_one('span').extract()

        novel = Novel(
            title=soup.select_one('.book-info h1').text.strip(),
            thumbnail_url=soup.select_one('.book-img img')['src'],
            synopsis=[t.strip() for t in synopsis_element.text.splitlines() if t.strip()],
            url=url,
        )

        for span in soup.select('.booktag span.blue'):
            for text in span.find_all(text=True, recursive=False):
                _text = str(text).strip()
                if _text.startswith('Author：'):
                    novel.author = _text.lstrip('Author：')
                    break

        for a in soup.select('.booktag a[href*="genre"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for i, a in enumerate(soup.select('#morelist > ul li a')):
            chapter = Chapter(
                index=i,
                title=a.text.strip(),
                url=self.base_urls[0] + a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.entry-content')
        self.clean_contents(content)

        content.unwrap()
        content = '<p>' + re.sub(r'<br ?/?>', '</p><p>', str(content).strip()) + '</p>'

        chapter.title = soup.select_one('.single-header h1').text.strip()
        chapter.paragraphs = content



import datetime
from urllib.parse import parse_qs, urlparse

from bs4 import BeautifulSoup

from .source import Source
from ...models import Chapter, Volume, Novel, Metadata


class WuxiaSite(Source):
    name = 'WuxiaWorld.site'
    base_urls = ('https://wuxiaworld.site',)
    last_updated = datetime.date(2021, 9, 3)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        title = ' '.join([
            str(x)
            for x in soup.select_one('.post-title').select_one('h1, h2, h3').contents
            if not x.name
        ]).strip()

        author = soup.select('.author-content a')
        if len(author) == 2:
            author = author[0].text + ' (' + author[1].text + ')'
        else:
            author = author[0].text

        novel = Novel(
            title=title,
            author=author,
            thumbnail_url=soup.select_one('.summary_image a img')['srcset'].split(' ', maxsplit=1)[0],
            synopsis=[str(p).strip() for p in soup.select_one('.summary__content > p:nth-child(2)').find_all(text=True, recursive=False)],
            url=url,
        )

        for a in soup.select('.artist-content > a'):
            novel.metadata.append(Metadata('contributor', a.text.strip(), others={'role': 'illustrator'}))

        for a in soup.select('a[href*="genre"][rel="tag"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        shortlink = soup.select_one('link[rel="shortlink"]')['href']
        novel_id = int(parse_qs(urlparse(shortlink).query).get('p')[0])
        response = self.request(
            'POST',
            'https://wuxiaworld.site/wp-admin/admin-ajax.php',
            data={
                'action': 'manga_get_chapters',
                'manga': novel_id,
            }
        )

        soup = BeautifulSoup(response.content, 'lxml')

        volume = Volume.default()
        novel.volumes.append(volume)
        for a in reversed(soup.select('li.wp-manga-chapter a')):
            index = len(volume.chapters)
            chapter = Chapter(
                index=index,
                title=a.text.strip() or f'Chapter {index}',
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        contents = soup.select('.text-left p, .cha-words p')
        body = [str(p) for p in contents if p.text.strip()]

        chapter.paragraphs = ''.join(body)



from novelsave_sources.sources import sources, meta_sources
from novelsave_sources.exceptions import UnknownSourceException


def parse_source(url):
    """Locate and create a source parser for the url if schema exists

    :throws UnknownSourceException: if the url cannot be parsed by any existing source schema
    """
    for source in sources:
        if source.of(url):
            return source()

    raise UnknownSourceException(f'"{url}" does not correspond to any available source')


def parse_metasource(url: str):
    """Locate and create a source parser for the url if schema exists

    :throws UnknownSourceException: if the url cannot be parsed by any existing source schema
    """
    for source in meta_sources:
        if source.of(url):
            return source()

    raise UnknownSourceException(f'"{url}" does not correspond to any available metadata source')


import datetime
from urllib.parse import urlparse, parse_qs

from bs4 import BeautifulSoup

from .source import Source
from ...models import Chapter, Novel, Metadata


class NovelSite(Source):
    base_urls = ('https://novelsite.net',)
    last_updated = datetime.date(2021, 9, 6)

    def __init__(self):
        super(NovelSite, self).__init__()
        self.bad_tags.extend([
            'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
        ])

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        authors = [a.text.strip() for a in soup.select('.author-content > a')]
        if len(authors) == 2:
            author = f'{authors[0]} ({authors[1]})'
        else:
            author = ', '.join(authors)

        title = soup.select_one('.post-title > h1').text.strip()
        if title.endswith(' Novel'):
            title = title[:-len(' Novel')]

        novel = Novel(
            title=title,
            author=author,
            thumbnail_url=soup.select_one('.summary_image img')['src'],
            synopsis=[p.text.strip() for p in soup.select('.summary__content > hr ~ p')],
            url=url,
        )

        # other metadata
        for item in soup.select('.post-content_item'):
            if item.select_one('.summary-heading').text.strip() == 'Alternative':
                for alt in item.select_one('.summary-content').text.strip().split(', '):
                    if alt != novel.title:
                        novel.metadata.append(Metadata('title', alt, others={'role': 'alt'}))
                break

        for artist in soup.select('.artist-content a'):
            novel.metadata.append(Metadata('contributor', artist.text.strip(), others={'role': 'ill'}))

        for genre in soup.select('.genres-content > a'):
            novel.metadata.append(Metadata('subject', genre.text.strip()))

        short_link = soup.select_one('link[rel="shortlink"]')['href']
        novel_id = parse_qs(urlparse(short_link).query).get('p')[0]
        response = self.session.post(
            'https://novelsite.net/wp-admin/admin-ajax.php',
            data={
                'action': 'manga_get_chapters',
                'manga': novel_id,
            }
        )

        soup = BeautifulSoup(response.content, 'lxml')

        volume = novel.get_default_volume()
        for a in reversed(soup.select('.wp-manga-chapter > a')):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.reading-content')

        # text-left has a better collection of paragraphs...
        # however we are not taking any chances assuming its always there
        text_left = content.select_one('.text-left')
        if text_left:
            content = text_left

        self.clean_contents(content)

        chapter.title = soup.select_one('.breadcrumb > li.active').text.strip(),
        chapter.paragraphs = str(content)



import re
from abc import ABC, abstractmethod
from typing import Tuple, Union
from urllib.parse import urlparse

from requests.cookies import RequestsCookieJar

from ..crawler import Crawler
from ...exceptions import UnavailableException
from ...models import Novel, Chapter


class Source(Crawler, ABC):
    name: str
    lang = 'en'
    login_viable: bool = False
    search_viable: bool = False

    @classmethod
    def of(cls, url: str) -> bool:
        """
        :param url: url to test
        :return: whether the url is from this source
        """
        return any(url.startswith(base_url) for base_url in cls.base_urls)

    def __init__(self):
        super(Source, self).__init__()

        # set default cookie domains
        if not hasattr(self, 'cookie_domains'):
            self.cookie_domains = []
            for url in self.base_urls:
                netloc = urlparse(url).netloc
                self.cookie_domains += [
                    netloc,
                    re.search(r'.+?(\..+)', netloc).group(1),  # remove the segment before first dot
                ]

    def login(self, email: str, password: str):
        """Login to the source and assign the required cookies"""
        raise UnavailableException(f"'{self.__name__}' scraper does not provide login functionality.")

    def search(self, keyword: str, *args, **kwargs):
        """Search for a novel on the source"""
        raise UnavailableException(f"'{self.__name__}' scraper does not provide search functionality.")

    def set_cookies(self, cookies: Union[RequestsCookieJar, Tuple[dict]]):
        """
        Replaces current cookiejar with given cookies

        :param cookies: new cookiejar
        """
        if type(cookies) == RequestsCookieJar:
            super(Source, self).set_cookies(cookies)
        elif type(cookies) == tuple:
            # clear preexisting cookies associated with source
            for domain in self.cookie_domains:
                try:
                    self.session.cookies.clear(domain=domain)
                except KeyError:
                    pass

            # add the dict formatted cookies
            for cookie in cookies:
                self.session.cookies.set(**cookie)
        else:
            raise TypeError(
                f"Unexpected type received: {type(cookies)}; Require either 'RequestsCookieJar' or 'Tuple[dict]'")

    @abstractmethod
    def novel(self, url: str) -> Novel:
        """Download and parse novel information

        :param url: link to novel profile
        :return: novel object containing volumes and metadata
        """
        raise NotImplementedError

    @abstractmethod
    def chapter(self, chapter: Chapter):
        """Download and parse chapter content

        Replaces the existing chapter's paragraphs attribute
        """
        raise NotImplementedError



from dataclasses import dataclass, field
from typing import Optional, List

from .metadata import Metadata
from .volume import Volume


@dataclass
class Novel:
    title: str
    url: str
    author: Optional[str] = None
    synopsis: List[str] = field(default_factory=lambda: [])
    thumbnail_url: str = None
    lang: str = 'en'

    volumes: List[Volume] = field(default_factory=lambda: [])
    metadata: List[Metadata] = field(default_factory=lambda: [])

    def get_default_volume(self):
        if self.volumes:
            return self.volumes[0]

        volume = Volume.default()
        self.volumes.append(volume)
        return volume



import datetime

from bs4 import BeautifulSoup

from .source import Source
from ...models import Novel, Volume, Chapter, Metadata


class ReadLightNovelsNet(Source):
    name = 'Read Light Novels'
    base_urls = ('https://readlightnovels.net',)
    last_updated = datetime.date(2021, 9, 6)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        authors = [a.text.strip() for a in soup.select('.info a[href*="novel-author"]')]
        if len(authors) == 2:
            author = f'{authors[0]} ({authors[1]})'
        else:
            author = ', '.join(authors)

        title = soup.select_one('.title').text.strip()
        if title.endswith(' Novel'):
            title = title[:-len(' Novel')]

        novel = Novel(
            title=title,
            author=author,
            synopsis=[p.text.strip() for p in soup.select('.desc-text > p')],
            thumbnail_url=soup.select_one('.info-holder img')['src'],
            url=url
        )

        for a in soup.select('a[rel*="tag"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        pages = soup.select('#pagination > ul > li:not(.dropup) a:last-child')
        pages_count = int(pages[-1]['title']) if pages else 0
        novel_id = soup.select_one('#id_post')['value']
        volume = novel.get_default_volume()

        for page_index in range(pages_count + 1):
            self.chapter_page(volume, novel_id, page_index + 1)

        return novel

    def chapter_page(self, volume: Volume, novel_id, page):
        response = self.session.post(
            'https://readlightnovels.net/wp-admin/admin-ajax.php',
            data={
                'action': 'tw_ajax',
                'type': 'pagination',
                'id': novel_id,
                'page': page,
            },
        )

        soup = BeautifulSoup(response.json()['list_chap'], 'lxml')
        for a in soup.select('ul.list-chapter li a'):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.chapter-content')
        self.clean_contents(content)
        for br in content.select('br'):
            br.extract()

        chapter.paragraphs = str(content)



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class WuxiaCom(Source):
    name = 'WuxiaWorld.com'
    base_urls = ('https://www.wuxiaworld.com',)
    last_updated = datetime.date(2021, 9, 4)

    blacklist_patterns = [
        r'^<span>(...|\u2026)</span>$',
        r'^translat(ed by|or)',
        r'(volume|chapter) .?\d+',
    ]

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.section-content h2').text,
            synopsis=[p.text.strip() for p in soup.select('.fr-view:not(.pt-10) p')],
            thumbnail_url=soup.select_one('img.media-object').get('src'),
            url=url,
        )

        author_elements = soup.select('.novel-body :is(dt, dd)')
        for i, element in enumerate(author_elements):
            if element.name != 'dt':
                continue

            head = element.text.strip()
            if head == 'Translator:':
                novel.metadata.append(Metadata('contributor', author_elements[i + 1].text.strip()))
            elif head == 'Author:':
                novel.author = author_elements[i + 1].text.strip()

        for a in soup.select('.genres a'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for panel in soup.select('#accordion .panel-default'):
            for a in panel.select('ul.list-chapters li.chapter-item a'):
                chapter = Chapter(
                    index=len(volume.chapters),
                    title=a.text.strip(),
                    url=self.base_urls[0] + a['href'],
                )

                volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        contents = soup.select_one('#chapterContent')
        if not contents:
            contents = soup.select_one('#chapter-content')
        if not contents:
            contents = soup.select_one('.panel-default .fr-view')
        if not contents:
            raise ValueError(f'Unable to read chapter content from {chapter}')

        for nav in (contents.select('.chapter-nav') or []):
            nav.extract()

        self.clean_contents(contents)

        chapter.title = soup.select_one('#chapter-outer  h4').text.strip()
        chapter.paragraphs = str(contents)



import datetime

from .source import Source
from ...models import Novel, Chapter, Metadata


class NovelOnlineFull(Source):
    base_urls = ('https://novelonlinefull.com',)
    last_updated = datetime.date(2021, 9, 7)

    blacklist_patterns = [
        r'^[\W\D]*(volume|chapter)[\W\D]+\d+[\W\D]*$',
        r'^(\s| )+$',  # non-breaking whitespace
    ]

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        synopsis_parent = soup.select_one('#noidungm')
        [e.extract() for e in synopsis_parent.select('h2, [style*="color"]')]

        novel = Novel(
            title=soup.select_one('.truyen_info_wrapper h1').text,
            thumbnail_url=soup.select_one('.info_image img')['src'],
            synopsis=self.find_paragraphs(synopsis_parent),
            url=url,
        )

        alternative = soup.select_one('.truyen_info_wrapper .truyen_info_right > li:first-child > span')
        if alternative:
            novel.metadata.append(Metadata('author', alternative.text.strip('Alternative :').strip(), others={'role': 'alternative'}))

        for li in soup.select('.truyen_info_wrapper .truyen_info_right > li'):
            span = li.select_one('span')
            if not span:
                continue

            label = span.text.strip()
            if label == 'Author(s):':
                novel.author = ', '.join([a.text.strip() for a in li.select('a')])
            elif label == 'GENRES:':
                for a in li.select('a'):
                    novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for i, a in enumerate(reversed(soup.select('.chapter-list > .row a'))):
            chapter = Chapter(
                index=i,
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('#vung_doc')
        self.clean_contents(content)

        chapter.paragraphs = str(content)



import datetime
import re
from typing import List
from urllib.parse import urlparse, parse_qs

from .source import Source
from ...models import Novel, Chapter, Metadata


class CreativeNovels(Source):
    name = 'Creative Novels'
    base_urls = ('https://creativenovels.com',)
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.x-bar-container > [class*="12"]').text.strip(),
            author=soup.select_one('.x-bar-container > [class*="14"]').text.strip().strip('Author: '),
            thumbnail_url=soup.select_one('img.book_cover')['src'],
            synopsis=[p.text.strip() for p in soup.select('.novel_page_synopsis > p')],
            url=url,
        )

        for a in soup.select('.suggest_tag > a'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        novel_id = parse_qs(urlparse(soup.select_one('link[rel="shortlink"]')['href']).query)['p'][0]
        security_key = ''
        for script in soup.select('script'):
            text = script.string
            if not text or 'var chapter_list_summon' not in text:
                continue

            p = re.findall(r'"([^"]+)"', text)
            if p[0] == 'ajaxurl' and p[1] == r'https:\/\/creativenovels.com\/wp-admin\/admin-ajax.php':
                if p[2] == 'security':
                    security_key = p[3]

        response = self.session.post(
            'https://creativenovels.com/wp-admin/admin-ajax.php',
            data=dict(
                action='crn_chapter_list',
                view_id=novel_id,
                s=security_key
            )
        )

        volume = novel.get_default_volume()
        volume.chapters = self.parse_chapter_list(response.content.decode('utf-8'))

        return novel

    def parse_chapter_list(self, content: str) -> List[Chapter]:
        chapters = []

        if not content.startswith('success'):
            return chapters

        content = content[len('success.define.'):]
        for data in content.split('.end_data.'):
            parts = data.split('.data.')
            if len(parts) < 2:
                continue

            chapter = Chapter(
                index=len(chapters),
                title=parts[1],
                url=parts[0],
            )

            chapters.append(chapter)

        return chapters

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('article .entry-content')
        for tag in content.select('.announcements_crn, .support-placement, span[style*="color:transparent"]'):
            tag.extract()

        self.clean_contents(content)

        chapter.title = soup.select_one('.entry-title').text.strip()
        chapter.paragraphs = str(content)



import datetime
from typing import Tuple, List

from .source import Source
from ...models import Novel, Chapter, Metadata


class DummyNovels(Source):
    name = 'Dummy novels'
    base_urls = ('https://dummynovels.com',)
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Tuple[Novel, List[Chapter], List[Metadata]]:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('h1.elementor-heading-title').text.strip(),
            thumbnail_url=soup.select_one('.elementor-image img')['src'],
            synopsis=[e.text.strip() for e in soup.select('.novel-synopsis-content > p') if e.text.strip()],
            url=url,
        )

        for element in soup.select('.elementor-widget-text-editor'):
            text = element.text.strip()
            if text.startswith('Author: '):
                novel.author = text.lstrip('Author: ')
            elif text.startswith('Translator: '):
                novel.metadata.append(Metadata('contributor', text.strip('Translator: '), others={'role': 'translator'}))
            elif text.startswith('Editors: '):
                novel.metadata.append(Metadata('contributor', text.strip('Editors: '), others={'role': 'editor'}))

        for element in soup.select('.novel-term > a'):
            novel.metadata.append(Metadata('subject', element.text.strip()))

        volume = novel.get_default_volume()
        for element in soup.select('.elementor-tab-content a[href*="novel"]:not(.elementor-accordion-title)'):

            # this removes the text '(NEW)'
            highlight = element.select_one('.new-highlight')
            if highlight:
                highlight.extract()

            chapter = Chapter(
                index=len(volume.chapters),
                title=element.text.strip(),
                url=element['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.elementor-widget-theme-post-content > div')

        # removes ads
        for element in content.select('.code-block'):
            element.extract()

        self.clean_contents(content)

        chapter.title = soup.select_one('.chapter-heading:not(.novel-title-for-chapters)').text.strip()
        chapter.paragraphs = str(content)



from setuptools import setup, find_packages

requirements = []
with open('requirements/common.txt', 'r') as f:
    requirements += f.readlines()
with open('requirements/prod.txt', 'r') as f:
    requirements += [l for l in f.readlines() if not l.startswith('-r')]

setup(
    install_requires=requirements,
    packages=find_packages(),
)



import datetime
import re

from .source import Source
from ...models import Novel, Chapter, Metadata


class LightNovelWorld(Source):
    base_urls = ('https://www.lightnovelworld.com',)
    last_updated = datetime.date(2021, 9, 7)

    chapter_title_regex = re.compile(r'chapter (\d+\.?\d*):? ?(.+)?', flags=re.IGNORECASE)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.novel-title').text.strip(),
            author=soup.select_one('.author a').text.strip(),
            thumbnail_url=soup.select_one('figure.cover img')['src'],
            synopsis=self.find_paragraphs(soup.select_one('.summary .content'), recursive=True),
            url=url,
        )

        for a in soup.select('.categories a'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        for a in soup.select('.tags a'):
            novel.metadata.append(Metadata('tag', a.text.strip()))

        novel.metadata.append(Metadata('status', soup.select_one('.completed, .ongoing').text.strip()))

        self.parse_toc(novel)

        return novel

    def parse_toc(self, novel):
        volume = novel.get_default_volume()

        # first page
        url = novel.url.rstrip("/") + '/chapters'
        soup = self.get_soup(url)

        last_page = soup.select_one('.PagedList-skipToLast a') or soup.select('.pagination li:not([class]) a')[-1]
        page_count = int(last_page['href'].rsplit('-', maxsplit=1)[-1])

        # get the rest of the pages
        for page in range(1, page_count + 1):
            url = f'{novel.url.rstrip("/")}/chapters/page-{page}'
            soup = self.get_soup(url)

            self.parse_toc_page(soup, volume)

    def parse_toc_page(self, soup, volume):
        for li in soup.select('.chapter-list li[data-chapterno]'):
            a = li.select_one('a')
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.select_one('strong').text.strip(),
                url=self.to_absolute_url(a['href']),
            )

            volume.chapters.append(chapter)

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        title_element = soup.find('div', {'class': 'titles'}).find('h2')
        result = self.chapter_title_regex.search(title_element.text)
        if len(result.groups()) > 1:
            chapter.title = result.group(2)
        else:
            chapter.title = title_element.text.strip()

        content = soup.select('.chapter-content > p:not([class])')
        chapter.paragraphs = ''.join(str(p) for p in content)



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class NovelFull(Source):
    name = 'NovelFull'
    base_urls = ('https://novelfull.com',)
    last_updated = datetime.date(2021, 9, 7)

    bad_tags = [
        'noscript', 'script', 'iframe', 'form', 'hr', 'img', 'ins',
        'button', 'input', 'amp-auto-ads', 'pirate',
        'h1', 'h2', 'h3'
    ]

    blacklist_patterns = [
        r'^\s*Translator:',
        r'^\s*Editor:',
        r'^\s*Atlas Studios',
        r'Read more chapter on NovelFull',
        r'full thich ung',
        r'If you find any errors \( broken links.*let us know < report chapter >',
    ]

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        image_element = soup.select_one('.info-holder .book img')

        authors = [a.text.strip() for a in soup.select('.info a[href*="/author"]')]
        if len(authors) == 2:
            author = f'{authors[0]} ({authors[1]})'
        else:
            author = ', '.join(authors)

        novel = Novel(
            title=image_element['alt'],
            author=author,
            thumbnail_url=self.base_urls[0] + image_element['src'],
            synopsis=[p.text.strip() for p in soup.select('.desc-text > p')],
            url=url,
        )

        for a in soup.select('.info a[href*="/genre"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        soup.select_one('.info').smooth()
        for div in soup.select('.info > div'):
            heading = div.select_one('h3').text.strip()
            value = div.find(text=True, recursive=False)
            if heading == 'Alternative names:':
                titles = value.split(', ')
                for title in titles:
                    novel.metadata.append(Metadata('title', title, others={'role': 'alt'}))
            elif heading == 'Source:':
                novel.metadata.append(Metadata('publisher', value))
            elif heading == 'Status:':
                novel.metadata.append(Metadata('status', div.select_one('a').text.strip()))

        last_pagination = soup.select_one('#list-chapter .pagination .last a')
        page_count = int(
            last_pagination['data-page']) if last_pagination else 0

        volume = novel.get_default_volume()
        for page in range(1, page_count + 2):
            self.parse_chapter_list(volume, url, page)

        return novel

    def parse_chapter_list(self, volume, novel_url, page):
        url = f'{novel_url.rstrip("/")}?page={page}&per-page=50'
        soup = self.get_soup(url)

        for a in soup.select('ul.list-chapter li a'):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a['title'].strip(),
                url=self.base_urls[0] + a['href'],
            )

            volume.chapters.append(chapter)

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('div#chapter-content')

        self.clean_contents(content)
        for ads in content.select('h3, h2, .adsbygoogle, script, ins, .ads, .ads-holder'):
            ads.extract()

        chapter.title = soup.select_one('.chapter-text').find(text=True, recursive=False).strip()
        chapter.paragraphs = str(content)



import datetime
from typing import List, Tuple

from .source import Source
from ...models import Chapter, Volume, Novel, Metadata


class NovelGate(Source):
    base_urls = ('https://novelgate.net',)
    last_updated = datetime.date(2021, 9, 3)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url.rstrip('/') + '/')

        film_content = soup.select_one('.film-content')
        for element in film_content.select('h3, .tags'):
            element.extract()

        novel = Novel(
            title=soup.select_one('.film-info .name').text.strip(),
            author=soup.select_one('a[href*="author"]').text.strip(),
            thumbnail_url=soup.select_one('.film-info .book-cover')['data-original'],
            synopsis=[p.strip() for p in film_content.find_all(text=True, recursive=True) if p.strip()],
            url=url,
        )

        for a in soup.select('.film-info a[href*="genre"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        chapter_count = 0
        for i, div in enumerate(soup.select('#list-chapters > .book')):
            volume = Volume(i, div.select_one('.title a').text.strip())

            for a in div.select('.list-chapters > li a'):
                chapter = Chapter(
                    index=chapter_count,
                    title=a.text.strip(),
                    url=self.base_urls[0] + a['href'],
                )

                chapter_count += 1
                volume.chapters.append(chapter)

            novel.volumes.append(volume)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        contents = soup.select_one('#chapter-body')
        self.clean_contents(contents)

        for element in contents.select('.language'):
            element.extract()

        chapter.paragraphs = str(contents)



from dataclasses import dataclass

DUBLIN_CORE_TAGS = [
    'title', 'language', 'subject',
    'creator', 'contributor', 'publisher', 'rights',
    'coverage', 'date', 'description',
]


@dataclass
class Metadata:
    """
    'DC' also known as Dublin Core metadata contains the following:

    Minimal required:

    - DC:identifier
    - DC:title
    - DC:language

    Optional:

    - DC:creator
    - DC:contributor
    - DC:publisher
    - DC:rights
    - DC:coverage
    - DC:date
    - DC:description
    """

    DUBLIN_CORE = 'DC'
    CUSTOM = 'OPF'

    name: str
    value: str
    others: dict
    namespace: str

    def __init__(self, name: str, value: str, others: dict = None):
        self.name = name
        self.value = value
        self.others = others or {}
        self.namespace = self.DUBLIN_CORE if name in DUBLIN_CORE_TAGS else self.CUSTOM



import datetime
from typing import List

import requests
from bs4 import BeautifulSoup

from .source import Source
from ...models import Chapter, Novel, Metadata


class ScribbleHub(Source):
    name = 'Scribble Hub'
    base_urls = ('https://www.scribblehub.com',)
    last_updated = datetime.date(2021, 9, 6)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('div.fic_title').text.strip(),
            author=soup.select_one('span.auth_name_fic').text.strip(),
            synopsis=[p.text.strip() for p in soup.select('.wi_fic_desc > p')],
            thumbnail_url=soup.select_one('.fic_image img')['src'],
            url=url,
        )

        for a in soup.select('a.fic_genre'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        for a in soup.select('a.stag'):
            novel.metadata.append(Metadata('tag', a.text.strip()))

        id_ = int(url.split('/')[4])

        volume = novel.get_default_volume()
        volume.chapters = self.parse_toc(id_)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        chapter.title = soup.select_one('.chapter-title').text.strip()
        chapter.paragraphs = str(soup.select_one('#chp_raw'))

    @staticmethod
    def parse_toc(id_: int) -> List[Chapter]:
        response = requests.post(
            'https://www.scribblehub.com/wp-admin/admin-ajax.php',
            data={
                'action': 'wi_gettocchp',
                'strSID': id_,
            },
        )

        soup = BeautifulSoup(response.content, 'lxml')

        chapters = []
        for i, a in enumerate(reversed(soup.select('li > a'))):
            chapter = Chapter(
                index=i,
                title=a.text.strip(),
                url=a['href']
            )

            chapters.append(chapter)

        return chapters



import datetime
import re
import unicodedata

from .source import Source
from ...models import Novel, Chapter, Metadata


class ReadLightNovel(Source):
    name = 'Read Light Novel'
    base_urls = ('https://www.readlightnovel.me',)
    last_updated = datetime.date(2021, 9, 7)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.block-title h1').text.strip(),
            author=soup.select_one("a[href*=author]").text.strip(),
            synopsis=[p.text.strip() for p in soup.select('.novel-right .novel-detail-item .novel-detail-body > p')],
            thumbnail_url=soup.select_one('.novel-cover img')['src'],
            url=url
        )

        for a in soup.select('.novel-detail-item.color-gray a'):
            novel.metadata.append(Metadata('title', a.text.strip(), others={'role': 'alt'}))

        for a in soup.select('.novel-details a[href*="category"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        for a in soup.select('.novel-details a[href*="tags"]'):
            novel.metadata.append(Metadata('tag', a.text.strip()))

        volume = novel.get_default_volume()
        for a in soup.select('.chapters .chapter-chs li a'):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.find('div', {'class': 'chapter-content3'}).find('div', {'class': 'desc'})

        title_element = content.find('h3')

        title = None
        if title_element is not None:
            # title can have some unicode characters in them
            title = unicodedata.normalize("NFKD", title_element.text)

        # paragraphs
        paragraphs = []

        if content.find('p') is None:
            # no <p> elements
            for text in content.find_all(text=True, recursive=False):
                paragraphs.append(text)
        else:
            para_elements = content.find_all('p', recursive=False)
            for element in para_elements:
                text = element.text.strip()
                paragraphs.append(text)

        # fixing the formatting issue
        for i in range(len(paragraphs)):
            paragraphs[i] = re.sub(r' \. ?', '. ', paragraphs[i]).strip()

        chapter.title = title
        chapter.paragraphs = paragraphs



import datetime

from .source import Source
from ...models import Chapter, Novel, Metadata


class NovelPassion(Source):
    base_urls = ('https://www.novelpassion.com',)
    last_updated = datetime.date(2021, 9, 4)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        authors = [a.text.strip() for a in soup.select('.dns .stq[href*="author"]')]
        synopsis_paragraphs = [p.text.strip() for p in soup.select('.j_synopsis p')]

        novel = Novel(
            title=soup.select_one('.psn h2.ddm').text.strip(),
            author=', '.join(authors),
            thumbnail_url=self.base_urls[0] + soup.select_one('.g_thumb img')['src'],
            synopsis=synopsis_paragraphs,
            url=url,
        )

        for a in soup.select('.dns .stq[href*="category"]'):
            novel.metadata.append(Metadata('subject', a.text.strip()))

        volume = novel.get_default_volume()
        for i, a in enumerate(reversed(soup.select('.content-list a'))):
            chapter = Chapter(
                index=i,
                title=a.select_one('span').text.strip(),
                url=self.base_urls[0] + a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.cha-words')
        self.clean_contents(content)

        chapter.title = soup.select_one('h2.dac').text.strip()
        chapter.paragraphs = str(content)



import datetime
from typing import List

from .source import Source
from ...models import Novel, Chapter, Metadata


class RainOfSnow(Source):
    name = 'Rain Of Snow Translations'
    base_urls = ('https://rainofsnow.com/',)
    last_updated = datetime.date(2021, 9, 4)

    def novel(self, url: str) -> Novel:
        soup = self.get_soup(url)

        novel = Novel(
            title=soup.select_one('.text h2').text.strip(),
            synopsis=self.synopsis(soup),
            thumbnail_url=soup.select_one('.imagboca1 img')['data-src'],
            url=url,
        )

        for li in soup.select('.vbtcolor1 > li'):
            key, value = li.select_one('.vt1').text.strip(), li.select_one('.vt2').text.strip()
            if key.lower() == 'author':
                novel.author = value

            elif key.lower() == 'translator':
                novel.metadata.append(Metadata('contributor', value, others={'role': 'translator'}))

            elif key.lower() == 'editor':
                novel.metadata.append(Metadata('contributor', value, others={'role': 'editor'}))

            elif key.lower() == 'genre(s)':
                for word in value.split(','):
                    if not word.strip():
                        continue
                    novel.metadata.append(Metadata('subject', word.strip()))

            elif key.lower() == 'type':
                novel.metadata.append(Metadata('type', value))

        for tag in soup.select('.vbtcolor1 > li a[href*="tag"]'):
            novel.metadata.append(Metadata('tag', tag.strip()))

        volume = novel.get_default_volume()
        for a in soup.select('#chapter .chapter a'):
            chapter = Chapter(
                index=len(volume.chapters),
                title=a.text.strip(),
                url=a['href'],
            )

            volume.chapters.append(chapter)

        return novel

    @staticmethod
    def synopsis(soup) -> List[str]:
        paragraphs = []
        for element in soup.select('#synop > p'):
            if not element.text.strip():
                continue

            paragraphs.append(element.text)

        return paragraphs

    def chapter(self, chapter: Chapter):
        soup = self.get_soup(chapter.url)

        content = soup.select_one('.bb-item[style*="block"] .content .scroller .zoomdesc-cont')
        content = self.clean_contents(content)

        chapter.paragraphs = str(content)
